/*
 * context.c - context routines
 *
 * Copyright (C) 1997,1998,1999 by Dominik Madon
 *
 * This source file is distributed "as is" in the hope that it will be
 * useful.  The tool set comes with no warranty, and no author or
 * distributor accepts any responsibility for the consequences of its
 * use. 
 * 
 * Everyone is granted permission to copy, modify and redistribute
 * this tool set under the following conditions:
 * 
 *    This source code is distributed for non-commercial use only. 
 *    Please contact the maintainer for restrictions applying to 
 *    commercial use.
 *
 *    Permission is granted to anyone to make or distribute copies
 *    of this source code, either as received or modified, in any
 *    medium, provided that all copyright notices, permission and
 *    nonwarranty notices are preserved, and that the distributor
 *    grants the recipient permission for further redistribution as
 *    permitted by this document.
 *
 *    Permission is granted to distribute this file in compiled
 *    or executable form under the same conditions that apply for
 *    source code, provided that either:
 *
 *    A. it is accompanied by the corresponding machine-readable
 *       source code,
 *    B. it is accompanied by a written offer, with no time limit,
 *       to give anyone a machine-readable copy of the corresponding
 *       source code in return for reimbursement of the cost of
 *       distribution.  This written offer must permit verbatim
 *       duplication by anyone, or
 *    C. it is distributed by someone who received only the
 *       executable form, and is accompanied by a copy of the
 *       written offer of source code that they received concurrently.
 *
 * In other words, you are welcome to use, share and improve this
 * source file.  You are forbidden to forbid anyone else to use, share
 * and improve what you give them.
 *
 * INTERNET: dominik.madon@epfl.ch
 * Snail: LSL-DI/EPFL, In-Ecublens, 1015 Lausanne
 *
 * $Id: context.c,v 1.2 2002/01/04 03:14:38 yeung Exp $
 *
 *
 */

#include "machine.h"
#include "misc.h"
#include "context.h"
#include "host.h"

/* total number of simulated contexts */
SS_CONTEXT_TYPE context_num;
/* DY: context configuration specification */
int context_config[SS_NUM_CONTEXT] =
  {1, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 
   0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0, 0};
int context_nelt = 1;
int context_master[SS_NUM_CONTEXT];

/* message port unit */
struct msg_port_t msg_data[SS_NUM_CONTEXT];

/*
 * msg_port_init()
 *
 * Initialize msg_port structure.
 *
 */

void
msg_port_init(void)
{
  int i;

  for(i = 0; i < SS_NUM_CONTEXT; i++) {
    msg_data[i].read = FALSE;
    msg_data[i].wrote = FALSE;
    msg_data[i].done = FALSE;
  } /* for */

} /* msg_port_init() */




void
msg_dump(void)
{
  int i;

  fprintf(stderr, "+-----------+-----------+---+---+---+-----+-----+\n");
  fprintf(stderr, "| port      | value     | w | r | d | Sid | Rid |\n");
  fprintf(stderr, "+-----------+-----------+---+---+---+-----+-----+\n");
  for(i = 0; i < (SS_NUM_CONTEXT); i++) {
    if (msg_data[i].read || msg_data[i].wrote) {
      fprintf(stderr, "| %9d | %9d | %s | %s | %s |  %02d |  %02d |\n",
              msg_data[i].port, msg_data[i].value,
              (msg_data[i].wrote ? "t" : "f"), 
              (msg_data[i].read ? "t" : "f"),
              (msg_data[i].done ? "t" : "f"),
              msg_data[i].sender_id,
              msg_data[i].receiver_id);
      fprintf(stderr, "+-----------+-----------+---+---+---+-----+-----+\n");
    } /* if */
  } /* if */

} /* msg_dump */





/*
 * msg_port_write()
 *
 * This procedure write a message in the msg_data structure.
 * Returns 0 if nobody was listening this port.
 * Returns 1 if a port was listening by another context.
 * Returns 2 if the message was already delivered.
 *
 */

unsigned int
msg_port_write(
  SS_CONTEXT_TYPE context,
  qword_t port,
  qword_t value,
  SS_CONTEXT_TYPE * receiver_context)
{
  int i = 0;
  int freecell = SS_NUM_CONTEXT;

  while ( i < SS_NUM_CONTEXT) {
    if (msg_data[i].read && !msg_data[i].done &&
        (msg_data[i].port == port)) {
      msg_data[i].sender_id = context;
      msg_data[i].value = value;
      msg_data[i].done = TRUE;
      *receiver_context = msg_data[i].receiver_id;
      return 1;        
    } else if (!msg_data[i].read && !msg_data[i].wrote && 
               (freecell == SS_NUM_CONTEXT)) {
      freecell = i;
    } else if (msg_data[i].done && msg_data[i].wrote &&
        (msg_data[i].port == port) && (msg_data[i].sender_id == context)) {
      msg_data[i].done = FALSE;
      msg_data[i].wrote = FALSE;
      return 2;
    }
    i++;
  } /* while */

  msg_data[freecell].sender_id = context;
  msg_data[freecell].port = port;
  msg_data[freecell].value = value;
  msg_data[freecell].wrote = TRUE;
  *receiver_context = msg_data[freecell].receiver_id;
  return 0;               

} /* msg_port_write */



/*
 * msg_port_read()
 *
 * This function reads the given port message.
 * Returns 0 if nobody sent a message to that port.
 * Returns 1 if the sender was already waiting a message.
 * Returns 2 if this message was already delivered.
 *
 */

unsigned int
msg_port_read(
  SS_CONTEXT_TYPE context,
  qword_t port,
  qword_t * value,
  SS_CONTEXT_TYPE * sender_context)
{
  int i = 0;
  int freecell = SS_NUM_CONTEXT;

  while ( i < SS_NUM_CONTEXT) {
    if (msg_data[i].wrote && !msg_data[i].done &&
        (msg_data[i].port == port)) {
      msg_data[i].receiver_id = context;
      *value = msg_data[i].value;
      msg_data[i].done = TRUE;
      *sender_context = msg_data[i].sender_id;
      return 1;        
    } else if (!msg_data[i].read && !msg_data[i].wrote && 
               (freecell == SS_NUM_CONTEXT)) {
      freecell = i;
    } else if (msg_data[i].done && msg_data[i].read &&
        (msg_data[i].port == port) && (msg_data[i].receiver_id == context)) {
      msg_data[i].done = FALSE;
      msg_data[i].read = FALSE;
      *value = msg_data[i].value;
      return 2;
    }
    i++;
  } /* while */

  msg_data[freecell].receiver_id = context;
  msg_data[freecell].port = port;
  *value = msg_data[freecell].value;
  msg_data[freecell].read = TRUE;
  *sender_context = msg_data[freecell].sender_id;
  return 0;             
  
} /* msg_port_read */




/* 
 * port_used_by_context()
 *
 * returns 0 if current context is not wating for a message transfer
 *          or if it is stopped, or
 * returns the port number used by the context whose at the moment
 *          waits for data transmitting/receiving. 
 *
 */

unsigned int 
port_used_by_context(SS_CONTEXT_TYPE context)
{
  int i;
  
  for(i = 0; i < SS_NUM_CONTEXT; i++)
/*      if ((msg_data[i].read && !msg_data[i].done && */
/*           msg_data[i].receiver_id == context) || */
/*          (msg_data[i].wrote && !msg_data[i].done && */
/*           msg_data[i].sender_id == context)) { */
    if ((msg_data[i].read && msg_data[i].receiver_id == context) ||
        (msg_data[i].wrote && msg_data[i].sender_id == context)) {
      return msg_data[i].port;
    } /* if */
  return 0;

} /* port_used_by_context */




SS_CONTEXT_TYPE
context_using_port(qword_t port)
{
  int i;
  
  for(i = 0; i < SS_NUM_CONTEXT; i++)
    if ((msg_data[i].wrote && !msg_data[i].done &&
         msg_data[i].port == port)) {
      return msg_data[i].sender_id;
    } else if ((msg_data[i].read && !msg_data[i].done &&
         msg_data[i].port == port)) {
      return msg_data[i].receiver_id;
    } /* if */
  return 0;

} /* context_using_port */





void
msg_port_delete(SS_CONTEXT_TYPE context)
{
  int i;
  
  for(i = 0; i < SS_NUM_CONTEXT; i++)
    if ((msg_data[i].read && !msg_data[i].done &&
         msg_data[i].receiver_id == context) ||
        (msg_data[i].wrote && !msg_data[i].done &&
         msg_data[i].sender_id == context)) {
      msg_data[i].read = FALSE;
      msg_data[i].wrote = FALSE;
    } /* if */

} /* msg_port_delete */



/*
 * msg_port_test()
 *
 * returns 0 if there is no port used with the given number.
 * returns 1 if there is a waiting context for a data on that port number.
 * returns 2 if there is a waiting context that wrote a data to that port.
 *
 */

unsigned int
msg_port_test(qword_t port)
{
  int i = 0;

  while ( i < SS_NUM_CONTEXT) {
    if (msg_data[i].port == port)
      if (msg_data[i].read &&
          !msg_data[i].done)
        return 1;
      if (!msg_data[i].read &&
          msg_data[i].done)
        return 2;
      i++;
  } /* while */

  return 0;
} /* msg_port_test */






/* lock structure */
struct lock_port_t lock_data[2 * SS_NUM_CONTEXT];


void
lock_port_init(void)
{
  int i;

  for(i = 0; i < (2 * SS_NUM_CONTEXT); i++) {
    lock_data[i].port = 0;
    lock_data[i].free = TRUE;
  } /* for */

} /* lock_port_init */




void
lock_dump(void)
{
  int i;

  fprintf(stderr, "+------+------------+------------+----+----+\n");
  fprintf(stderr, "| lock | port       | next_port  |free|ctxt|\n");
  fprintf(stderr, "+------+------------+------------+----+----+\n");
  for(i = 0; i < (2 * SS_NUM_CONTEXT); i++) {
    if (!lock_data[i].free) {
      fprintf(stderr, "| 0x%02X |      %5d |      %5d |  %s | %2d |\n",
              i, lock_data[i].port, lock_data[i].next_port,
              (lock_data[i].free ? "t" : "f"), lock_data[i].context);
      fprintf(stderr, "+------+------------+------------+----+----+\n");
    } /* if */
  } /* if */

} /* lock_dump */




/*
 * lock_port_lock()
 *
 * This function tries to lock a "lock port". If the lock
 * is free, the given value is written to lock's value and
 * the free flag is cleared.
 *
 * If already locked, the lock's value is taken as
 * a port number whose is used to suspend (write in a port)
 * the caller. The value given by the caller is writtent in
 * the lock's value.
 *
 * returns 1 if lock was unlocked.
 * returns 0 if lock was already locked.
 *
 */

unsigned int
lock_port_lock(
  SS_CONTEXT_TYPE context, 
  qword_t lock_id, 
  qword_t value)
{
  if (lock_data[lock_id].free) {
    lock_data[lock_id].port = value;
    lock_data[lock_id].next_port = value;
    lock_data[lock_id].free = FALSE;
    lock_data[lock_id].context = context;
/*      lock_dump(); */
/*      msg_dump(); */
/*      fprintf(stderr, "#%d# nouvelle entre port:%d, value:%d\n", */
/*              context, lock_id, value); */
    return 1;
  } else if (lock_data[lock_id].context == context) {
    lock_data[lock_id].port = value;
/*      lock_dump(); */
/*      msg_dump(); */
/*      fprintf(stderr, "#%d# changement entre port:%d, value:%d\n", */
/*              context, lock_id, value); */
    return 1;    
  } else {
    SS_CONTEXT_TYPE receiver;

    if (!port_used_by_context(context)) {
/*        fprintf(stderr, "# context:%d, next_port:%d, value:%d\n", */
/*                context, lock_data[lock_id].next_port, value); */
      msg_port_write(context, lock_data[lock_id].next_port, value, &receiver);
      lock_data[lock_id].next_port = value;    
    } /* if */
/*      lock_dump(); */
/*      msg_dump(); */
/*      fprintf(stderr, "#%d# lock => data entre port:%d, value:%d\n", */
/*              context, lock_id, value); */
    return 0;
  } /* if */

} /* lock_port_lock */




/*
 * lock_port_unlock()
 *
 * returns 1 if the lock becomes free.
 * returns 0 if there is another context that will use
 * the lock.
 *
 */

unsigned int
lock_port_unlock(
  SS_CONTEXT_TYPE context, 
  qword_t lock_id, 
  SS_CONTEXT_TYPE * next_context)
{
  if (lock_data[lock_id].port != lock_data[lock_id].next_port) {
    *next_context = context_using_port(lock_data[lock_id].port);
    msg_port_delete(*next_context);
    lock_data[lock_id].context = *next_context;
/*      lock_dump(); */
/*      msg_dump(); */
    return 0;
  } else {
    lock_data[lock_id].port = 0;
    lock_data[lock_id].next_port = 0;
    lock_data[lock_id].free = TRUE;
/*      lock_dump(); */
/*      msg_dump(); */
    return 1;
  } /* if */

} /* lock_port_unlock */  


/*
 * context_changed()
 *
 * return 0 if nothing changed.
 * return -1 if context(s) have been freezed.
 * return 1 if context(s) have been woke up.
 *
 * Entries are not checked.
 *
 * If there is a mixed change -1 is returned. (FIXMEDM)
 *
 */

unsigned int
context_changed(
  qword_t old_reg,
  qword_t new_reg,
  SS_CONTEXT_TYPE *mask)
{
  if (old_reg == new_reg)
    return CTXT_CHNG_NONE;
  *mask = old_reg ^ new_reg;

  /* Prepare a variable to check if the context */
  /*  was woke up or freezed */

  old_reg = old_reg & *mask;

  /* Check if context was woke up or freezed */

  if (old_reg)
    return CTXT_CHNG_FREEZE;
  else
    return CTXT_CHNG_WAKEUP;

} /* context_changed */



/* the next list is used to store PC values of context's next
   instruction to execute */

struct rbackup_t rbackup_data[SS_NUM_CONTEXT];

