#ifndef SSMT_LIB_H
#define SSMT_LIB_H

#define NOREORDER(cmd) {\
  asm (".set	noreorder");\
  cmd;\
  asm (".set	reorder");\
}

#define VSIM_MAX_STREAM_DESC 2048
#define SS_NUM_CONTEXT 32

#define VSIM_ENTER 	 						\
	NOREORDER(asm volatile ("perr $0, $1, $0"))
#define VSIM_EXIT 	 						\
	NOREORDER(asm volatile ("perr $0, $2, $0"))
#define MARK1 	 	 						\
	NOREORDER(asm volatile ("perr $0, $3, $0"))
#define MARK2 	 	 						\
	NOREORDER(asm volatile ("perr $0, $4, $0"))
#define SIM_PRINTF(X) 	 						\
	NOREORDER(asm volatile ("perr %0, $5, $0" :: "r" (X)))
	
#define SSMT_PREFETCH(X) 						\
	NOREORDER(asm volatile ("perr %0, $0, $1" :: "r" (X)))
	
#define SSMT_RMSG(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $2" : "=r" (X) : "r" (Y)))
#define SSMT_SMSG(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $3" :: "r" (X), "r" (Y)))
#define SSMT_TMSG(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $4" : "=r" (X) : "r" (Y)))
#define SSMT_WCXT(X, Y, Z) 						\
	NOREORDER(asm volatile ("minub8 %0, %1, %2" :: "r" (X), "r" (Y), "r" (Z)))
#define SSMT_RCXT(X, Y, Z) 						\
	NOREORDER(asm volatile ("minsb8 %0, %1, %2" :: "r" (X), "r" (Y), "r" (Z)))
#define SSMT_LACXT(X) 							\
	NOREORDER(asm volatile ("perr %0, $0, $5" :: "r" (X)))
#define SSMT_SACXT(X) 							\
	NOREORDER(asm volatile ("perr %0, $0, $6" : "=r" (X)))
#define SSMT_LCXTCTRL(X) 						\
	NOREORDER(asm volatile ("perr %0, $0, $7" :: "r" (X)))
#define SSMT_SCXTCTRL(X) 						\
	NOREORDER(asm volatile ("perr %0, $0, $8" : "=r" (X)))
#define SSMT_LMMU(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $9" :: "r" (X), "r" (Y)))
#define SSMT_SMMU(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $10" : "=r" (X) : "r" (Y)))
#define SSMT_RPORB(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $11" : "=r" (X) : "r" (Y)))
#define SSMT_WPORB(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $12" :: "r" (X), "r" (Y)))
#define SSMT_LCU(X, Y) 							\
	NOREORDER(asm volatile ("perr %0, %1, $13" :: "r" (X), "r" (Y)))
#define SSMT_SCU(X, Y) 							\
	NOREORDER(asm volatile ("perr %0, %1, $14" : "=r" (X) : "r" (Y)))
#define SSMT_SCUA(X) 							\
	NOREORDER(asm volatile ("perr %0, $0, $15" : "=r" (X)))
#define SSMT_SCUB() 							\
	NOREORDER(asm volatile ("perr $0, $0, $16"))
#define SSMT_PMSG(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $17" :: "r" (X), "r" (Y)))
#define SSMT_GMSG(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $18" : "=r" (X) : "r" (Y)))
#define SSMT_LCK(X, Y) 							\
	NOREORDER(asm volatile ("perr %0, %1, $19" :: "r" (X), "r" (Y)))
#define SSMT_UNLCK(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $20" : "=r" (X) : "r" (Y)))
#define SSMT_LPCR(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $21" :: "r" (X), "r" (Y)))
#define SSMT_SPCS(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $22" : "=r" (X) : "r" (Y)))
#define SSMT_SPCSA(X, Y) 						\
	NOREORDER(asm volatile ("perr %0, %1, $23" : "=r" (X) : "r" (Y)))

#define SSMT_SEMS(CONTEXT, X)						\
	NOREORDER(asm volatile ("perr %0, %1, $24" :: "r" (CONTEXT), "r" (X)))
#define SSMT_SEMP(X)							\
	NOREORDER(asm volatile ("perr %0, $0, $25" :: "r" (X)))
#define SSMT_SEMV(X)							\
	NOREORDER(asm volatile ("perr %0, $0, $26" :: "r" (X)))
#define SSMT_SEMVP(X)							\
	NOREORDER(asm volatile ("perr %0, $0, $27" :: "r" (X)))


typedef unsigned int md_word_t;
typedef unsigned int SS_CONTEXT_TYPE;


#define loadMMUconfig(config, context)	SSMT_LMMU(config, context)
#define getMMUconfig(config, context)	SSMT_SMMU(config, context)
	
#define loadContextUnitReg(data, reg)	SSMT_LCU(data, reg)
#define getContextUnitReg(data, reg)	SSMT_SCU(data, reg)

#define putMessage(data, port)		SSMT_PMSG(data, port)
#define getMessage(data, port) 		SSMT_GMSG(data, port)

#define lock(port, lock)		SSMT_LCK(port, lock)
#define unlock(port, lock)		SSMT_ULCK(port, lock)

#define loadPCandResume(pc, context)	SSMT_LPCR(pc, context)
#define getPCandStop(pc, context)	SSMT_SPCS(pc, context)

#define createThread(threadId, context, stack, func, params) 		\
  createThread_(threadId, context, stack, func, params, 0x1000);

/* FIXME: CHOI, stack index should be modified, also stw or stq ? */
#define createThread_(threadId, context, stack, func, params, port) 	\
  ((unsigned *)stack)[-7] = (unsigned)threadId; 			\
  ((unsigned *)stack)[-6] = (unsigned)params; 				\
  ((unsigned *)stack)[-5] = (unsigned)func; 				\
  NOREORDER(asm volatile ("stw $29, -16(%0)" : : "r" (stack))); 		\
  SSMT_LPCR(_startThread, context);					\
  SSMT_PMSG(stack, port);						

#define waitThread(data, threadId)	SSMT_GMSG(data, threadId+0x1000)

#define activeContext_p(running, context)				\
{ 									\
  unsigned active;							\
  SSMT_SCU(active, 0);							\
  running = (1 << context) & active;					\
} 

#define suspendContext(pc, context) 	SSMT_SPCS(pc, context)

#define resumeContext(value) 						\
{ 									\
  unsigned cxt_PC; 							\
  SSMT_SPCS(cxt_PC, value);						\
  SSMT_LPCR(ctx_PC, value);						\
}

#if 0 /* CHOI: not used */

#define writePort(data, port)		SSMT_WPORT(data, port)
#define readPort(data, port)		SSMT_RPORT(data, port)

#define getRegACXT(value) 		SSMT_SCU(value, 0)
#define getCXT(value) 			SSMT_SCUA(value)
#define setSP() 			SSMT_SCUB()

int swapThreadOut(
      SS_CONTEXT_TYPE context, 
      md_word_t *sp, 
      md_word_t *pc);

int swapThreadIn(
      SS_CONTEXT_TYPE context,
      md_word_t sp,
      md_word_t pc);

extern void _startThread();

#endif

#endif
